class Either:
    """
    The Either type represents values with two possibilities: B value of type Either<A, B> is either Left<A> or Right<B>
    But not both in the same time.
    """
    def __init__(self, value):
        self.value = value


class Left(Either):

    def map(self, left_fn, right_fn):
        """
        takes 2 functions and applied first one on current Either value and returns new left Either mapped value
        :param left_fn: mapper for left value of Either
        :type (A) -> B
        :param right_fn: mapper for right value of Either, never called
        :type: (Any) -> Any
        :return: Left<B>
        """
        return Left(left_fn(self.value))

    def fold(self, left_fn, right_fn):
        return left_fn(self.value)


class Right(Either):

    def map(self, left_fn, right_fn):
        """
        takes 2 functions and applied second one on current Either value and returns new left Either mapped value
        :param left_fn: mapper for left value of Either, never called
        :type: (Any) -> Any
        :param right_fn: mapper for right value of Either
        :type (A) -> B
        :return: Left<B>
        """
        return Right(right_fn(self.value))

    def fold(self, left_fn, right_fn):
        return right_fn(self.value)


import pytest
from pymonet.lazy_box import LazyBox
from pymonet.utils import identity


class LazyBoxSpy:

    def fn(self):
        return 42

    def side_effect(self, input):
        return input


def test_lazy_box(mocker):
    """
    input function and all map functions should be called during called fold method
    """

    lazy_box_spy = LazyBoxSpy()
    mocker.spy(lazy_box_spy, 'fn')
    mocker.spy(lazy_box_spy, 'side_effect')

    lazy_box = LazyBox(lazy_box_spy.fn)
    assert lazy_box_spy.fn.call_count == 0

    lazy_box = lazy_box.map(lazy_box_spy.side_effect)
    assert lazy_box_spy.fn.call_count == 0
    assert lazy_box_spy.side_effect.call_count == 0

    lazy_box = lazy_box.map(lambda number: number + 1)
    assert lazy_box_spy.fn.call_count == 0
    assert lazy_box_spy.side_effect.call_count == 0

    assert lazy_box.fold(lambda number: number + 1) == 44
    assert lazy_box_spy.fn.call_count == 1
    assert lazy_box_spy.side_effect.call_count == 1




import pytest
from pymonet.either import Either, Left, Right
from pymonet.utils import identity


def either_operation(either):
    return either\
        .map(lambda value: 'prefix_{}'.format(value), lambda value: value + 1)\
        .fold(identity, identity)


def test_left_either():
    left = Left('some_string')
    result = either_operation(left)
    assert result == 'prefix_some_string'


def test_right_either():
    right = Right(42)
    result = either_operation(right)
    assert result == 43
